#include <debugging_regexis024/prettyprint/prettyprint_util.h>
#include <functional>
#include <libregexis024vm/utils.h>

namespace regexis024 {
    static const char* ch_empty = "    ";
    static const char* ch_passing_by = "\u2502   ";
    static const char* ch_connect_right_and_forward = "\u251c\u2500\u2500\u2500";
    static const char* ch_connect_right_last = "\u2514\u2500\u2500\u2500";

    static const char* ch_box_left_side = "\u2551";
    static const char* ch_box_right_side = "\u2551";
    static const char* ch_box_top_side = "\u2550";
    static const char* ch_box_bottom_side = "\u2550";
    static const char* ch_box_crn_top_left = "\u2554";
    static const char* ch_box_crn_top_right = "\u2557";
    static const char* ch_box_crn_bottom_left = "\u255A";
    static const char* ch_box_crn_bottom_right = "\u255D";

    size_t length_of_line(const std::string& str) {
        size_t ch = 0;
        size_t pos = 0;
        while (pos < str.size()) {
            int32_t code;
            size_t adj;
            utf8_string_iterat(code, adj, pos, str.data(), str.size());
            if (code < 0)
                return ch;
            ch++;
            pos += adj;
        }
        return ch;
    }

    /* Warning: recursion used */
    void toLines_dfs(const TreeWithStringsNode& node, lines& out, std::vector<bool>& prefix) {
        out.push_back("");
        size_t n = prefix.size();
        for (size_t i = 0; i < n; i++) {
            if (i + 1 < n) {
                out.back() += prefix[i] ? ch_passing_by : ch_empty;
            } else {
                out.back() += prefix[i] ? ch_connect_right_and_forward : ch_connect_right_last;
            }
        }
        out.back() += node.val;
        prefix.push_back(true);
        size_t m = node.childeren.size();
        for (size_t i = 0; i < m; i++) {
            if (i + 1 == m)
                prefix[n] = false;
            toLines_dfs(node.childeren[i], out, prefix);
        }
        prefix.pop_back();
    }

    void TreeWithStringsNode::toLines(lines &out) const {
        std::vector<bool> prefix;
        toLines_dfs(*this, out, prefix);
    }

    std::string strMul(size_t n, const char* str) {
        std::string res;
        for (size_t i = 0; i < n; i++)
            res += str;
        return res;
    }

    lines wrapWithBox(const lines &in) {
        lines out;
        size_t max_width = 0;
        for (auto& l: in)
            max_width = std::max(max_width, length_of_line(l));
        out.push_back(ch_box_crn_top_left + strMul(max_width, ch_box_top_side) + ch_box_crn_top_right);
        for (auto& line: in) {
            size_t s = length_of_line(line);
            out.push_back(ch_box_left_side + line + strMul(max_width - s, " ") + ch_box_right_side);
        }
        out.push_back(ch_box_crn_bottom_left + strMul(max_width, ch_box_bottom_side) + ch_box_crn_bottom_right);
        return out;
    }

    void printLines(const lines &in) {
        for (auto& l: in)
            printf("%s\n", l.c_str());
    }
}
