#ifndef LIBREGEXIS024_COLORED_CODESET_H
#define LIBREGEXIS024_COLORED_CODESET_H

#include <stdlib.h>
#include <stdint.h>
#include <vector>

#include <libregexis024fa/codeset.h>

namespace regexis024 {
    /* Used for determinizer. Nowhere else */
    struct ColoredCodesetSegment {
        uint32_t color;
        uint32_t right_code;
        ColoredCodesetSegment* next = NULL;

        /* Temporary varaible (used by apply_divisor() method) */
        bool divisor_on_left = false;

        ColoredCodesetSegment(uint32_t color, uint32_t right_code);
    };

    /* Warning!!! This stupid class is OOM-unsafe!!!
    * This is not an issue as far as you don't show any of it's instance to the user of libregexis024 */
    struct ColoredCodesetSegmentList {
        ColoredCodesetSegment* first = NULL;

        ColoredCodesetSegmentList();


        void replace_myself(const ColoredCodesetSegmentList& other);

        ColoredCodesetSegmentList(const ColoredCodesetSegmentList& other);

        /* Use only internally */
        void free_myself();

        ~ColoredCodesetSegmentList();

        ColoredCodesetSegmentList& operator=(const ColoredCodesetSegmentList& other);
    };

    /* Highly unoptimized algorithm on this data structure O(C^2) time*/
    class ColoredCodeset {
        ColoredCodesetSegmentList list;
        /* Size of this vector is equal to the number of colors */
        std::vector<std::vector<size_t>> requests;
        uint64_t DummyN;
        size_t nxt_request_id = 0;

        void split_phase(const codeset_t& X);
    public:
        /* First dummy_n split requests will be viewed as 'dummy requests', when complete map of splits is requested,
        * colors that are registed indide only dummy requests won't be returned. */
        ColoredCodeset(uint64_t dummy_n);

        /* O(C, which is bad, but my library's compiler is already slow by itself, so who cares) */
        void apply_divisor(const codeset_t& X);

        /* Returned 'requests' mapping will feature request id's with DummyN substituted from them */
        void get_splits_of_non_dummy(std::vector<codeset_t>& res_input,
            std::vector<std::vector<size_t>>& res_color_to_requests);
    };
}

#endif
