#include <libregexis024fa/fa_first_stage_fix.h>
#include <libregexis024fa/misc_fa_funcs.h>
#include <libregexis024vm/utils.h>
#include <assert.h>

// #ifdef LIBREGEXIS024_DEBUG
// #include <debugging_regexis024/debug_through_graphviz.h>
// #endif

namespace regexis024 {
    REGEX_IS024_FA_FirstStageFixInfo first_stage_fix_fa(FA_Container& sourceFa, FA_Container& resultFa) {
        assert(sourceFa.start);
        REGEX_IS024_FA_FirstStageFixInfo info;

        for (size_t I_scans = 0; I_scans < sourceFa.all.size(); I_scans++){
            FA_Node* beg = sourceFa.all[I_scans];
            if (beg->type != look_one_ahead)
                continue;
            FA_NodeOfLookOneAhead& loa = (*(FA_NodeOfLookOneAhead*)beg);
            codeset_t& restriction = loa.restriction;
            assert(loa.nxt_node);

            struct Marked{
                FA_Node* node;
                size_t refs_from_my = 1;
                bool making_copy = false;
                FA_Node* copy = NULL;

                explicit Marked(FA_Node *node) : node(node) {}
            };

            std::vector<Marked> searched;
            searched.emplace_back(loa.nxt_node);
            loa.nxt_node->search_mark = 0;

            for (size_t done = 0; done < searched.size(); done++){
                FA_Node& cur = *searched[done].node;
                for (FA_Node** nxtN : cur.get_all_empty_valid_transitions()){
                    if ((**nxtN).search_mark == -1){
                        assert((**nxtN).nodeId != loa.nodeId);
                        (**nxtN).search_mark = (int64_t)searched.size();
                        searched.emplace_back(*nxtN);
                    } else {
                        searched[(**nxtN).search_mark].refs_from_my++;
                    }
                }
            }
            std::vector<FA_Node*> s2s;
            for (auto& v_sete: searched){
                if (v_sete.refs_from_my < v_sete.node->refs){
                    v_sete.making_copy = true;
                    s2s.push_back(v_sete.node);
                }
            }
            while (!s2s.empty()){
                FA_Node& m = *s2s.back(); s2s.pop_back();
                assert(searched[m.search_mark].making_copy);
                /* Beacuse of this operation source Fa is not read-only. It becomes useless after renerating resultFa */
                searched[m.search_mark].copy = copy_fa_node(m, sourceFa);

                for (FA_Node** nxtN: m.get_all_empty_valid_transitions()){
                    Marked& nxtNaux = searched[(**nxtN).search_mark];
                    if (!nxtNaux.making_copy){
                        nxtNaux.making_copy = true;
                        s2s.push_back(*nxtN);
                    }
                }
            }

            for (auto& v_sete : searched){
                FA_Node* my = v_sete.making_copy ? v_sete.copy : v_sete.node;
                for (FA_Node** nxtN: my->get_all_empty_valid_transitions()){
                    Marked& nxtNaux = searched[(**nxtN).search_mark];
                    if (nxtNaux.making_copy)
                        reattach_fa_node_edge(nxtN, nxtNaux.copy);
                }
                my->apply_lookahead_restriction(restriction);
                if (my->type == match)
                    info.fed_chars_extend_one_right = true;
            }
            {
                Marked& loa_nxt_aux = searched[loa.nxt_node->search_mark];
                if (loa_nxt_aux.making_copy)
                    reattach_nxt_node(&loa, loa_nxt_aux.copy);
            }

            for (auto& v_sete: searched)
                v_sete.node->search_mark = -1;
        }

        // show_fa_with_sxiv_after_dot(sourceFa, {{}, {}}, {});

        {
            /* Now it's time to fill resultFa. Skipping all look one ahead's */
            auto skip_useless = [&](FA_Node* v) -> FA_Node* {
                while (v->type == look_one_ahead){
                    v = ((FA_NodeOfLookOneAhead*)v)->nxt_node;
                }
                return v;
            };

            resultFa.start = sourceFa.start;
            std::vector<FA_Node**> homework = {&(resultFa.start)};
            std::vector<FA_Node*> sourceIdToResNode(sourceFa.all.size(), NULL);

            while (!homework.empty()) {
                FA_Node** vPtr = homework.back(); homework.pop_back();
                FA_Node* right_source_v = skip_useless(*vPtr);
                size_t vid = right_source_v->nodeId;
                if (!sourceIdToResNode[vid]) {
                    sourceIdToResNode[vid] = copy_fa_node_to_another_fa(*right_source_v, resultFa);
                    for (FA_Node** uuPtr: sourceIdToResNode[vid]->get_all_transitions())
                        homework.push_back(uuPtr);
                }
                *vPtr = sourceIdToResNode[vid];
                sourceIdToResNode[vid]->refs++;
            }
        }


        {
            /* Guessing info.fed_chars_extend_one_left */
            size_t done = 0;
            std::vector<FA_Node*> searched;
            searched.push_back(resultFa.start);
            resultFa.start->search_mark = 0;
            while (done < searched.size()){
                if (searched[done]->type == look_one_behind){
                    info.fed_chars_extend_one_left = true;
                    break;
                }
                for (FA_Node** nxtN: searched[done]->get_all_empty_valid_transitions()){
                    if ((**nxtN).search_mark < 0){
                        (**nxtN).search_mark = 0;
                        searched.push_back(*nxtN);
                    }
                }
                done++;
            }
            for (FA_Node* d: searched)
                d->search_mark = -1;
        }
        return info;
    }

    FA_NodeOfOneCharRead* generate_alt_ending(const codeset_t& restriction, FA_Container& fa){
        FA_NodeOfOneCharRead* n1 = fa.makeOneCharRead(restriction, true);
        FA_NodeOfMatch* n2 = fa.makeMatch();
        n2->ext_filter_added = true;  // Won't actually be used
        reattach_fa_node_edge(&(n1->nxt_node), n2);
        return n1;
    }

    void regular_second_stage_fix(const FA_Container& sourceFa, FA_Container& resultFa,
                                  const REGEX_IS024_FA_FirstStageFixInfo &info1)
    {
        assert(resultFa.all.empty() && !resultFa.start);
        if (!sourceFa.start)
            return;
        resultFa.start = sourceFa.start;
        // A vector of pointers in resutFa to nodes that belong to sourceFa. They should undergo a little bit of copying.
        std::vector<FA_Node**> homework = {&(resultFa.start)};
        // source node id s index. Element is NULL if no copy (in resultFa) exists and resFa node if copying was performed
        std::vector<FA_Node*> sourceIdToResNode(sourceFa.all.size(), NULL);
        while (!homework.empty()) {
            FA_Node** vPtr = homework.back(); homework.pop_back();
            FA_Node* sourceV = *vPtr; assert(sourceV);
            size_t sourceVId = sourceV->nodeId;
            if (!sourceIdToResNode[sourceVId]) {
                if (sourceV->type == match) {
                    FA_NodeOfMatch& mn = dynamic_cast<FA_NodeOfMatch&>(*sourceV);
                    FA_NodeOfMatch* res_mn = resultFa.makeMatch();
                    if (info1.fed_chars_extend_one_right) {
                        FA_NodeOfOneCharRead* res_ocr2n = resultFa.makeOneCharRead(
                           mn.ext_filter_added ? mn.pending_filter : codeset_of_all, true);
                        reattach_nxt_node(res_ocr2n, res_mn);
                        sourceIdToResNode[sourceVId] = res_ocr2n;
                    } else {
                        sourceIdToResNode[sourceVId] = res_mn;
                    }
                } else {
                    sourceIdToResNode[sourceVId] = copy_fa_node_to_another_fa(*sourceV, resultFa);
                    /* O_o */
                    for (FA_Node** uuPtr: sourceIdToResNode[sourceVId]->get_all_transitions())
                        homework.push_back(uuPtr);
                }
            }
            *vPtr = sourceIdToResNode[sourceVId];
            sourceIdToResNode[sourceVId]->refs++;
        }

        if (info1.fed_chars_extend_one_left) {
            FA_NodeOfOneCharRead* ns = resultFa.makeOneCharRead(codeset_of_all, true);
            yay_new_start(resultFa, ns);
        }
    }
}
