#include <libregexis024fa/tracking_fa_nodes.h>
#include <assert.h>

namespace regexis024 {
    bool isImmMovOpcode(opcode_t inst) {
        return inst == opcodes::MOV_COLARR_IMM || inst == opcodes::MOV_SELARR_IMM;
    }

    bool isCurPosMovOpcode(opcode_t inst) {
        return inst == opcodes::MOV_COLARR_BTPOS || inst == opcodes::MOV_SELARR_CHPOS;
    }

    bool isColarrOpcode(opcode_t inst) {
        return inst == opcodes::MOV_COLARR_IMM || inst == opcodes::MOV_COLARR_BTPOS;
    }

    bool isSelarrOpcode(opcode_t inst) {
        return inst == opcodes::MOV_SELARR_IMM || inst == opcodes::MOV_SELARR_CHPOS;
    }

    bool isTrackingFaNode(const FA_Node *n) {
        return n->type == track_array_mov_imm || n->type == track_array_mov_halfinvariant;
    }

    std::string TrackingOperationInFa::toString() const {
        switch (opcode){
            case opcodes::MOV_COLARR_IMM:
                return "colarr[" + std::to_string(key) + "] := " + std::to_string(immValue);
            case opcodes::MOV_SELARR_IMM:
                return "selarr[" + std::to_string(key) + "] := " + std::to_string(immValue);
            case opcodes::MOV_COLARR_BTPOS:
                return "colarr[" + std::to_string(key) + "] := cur byte position";
            case opcodes::MOV_SELARR_CHPOS:
                return "selarr[" + std::to_string(key) + "] := cur char position";
            default:
                return "wrong collection operation";
        }
    }

    FA_NodePathPart* convert_to_node(const TrackingOperationInFa& op, FA_Container& fa) {
        if (isImmMovOpcode(op.opcode)) {
            return fa.makeTrackArrayMovImm(op.opcode, op.key, op.immValue);
        }
        assert(isCurPosMovOpcode(op.opcode));
        return fa.makeTrackArrayMovHalfinvariant(op.opcode, op.key);

    }
}
