#ifndef LIBREGEXIS024_SRC_LIBREGEXIS024SOL_EXPR_PARSE_FUNCTIONS_EPF_H
#define LIBREGEXIS024_SRC_LIBREGEXIS024SOL_EXPR_PARSE_FUNCTIONS_EPF_H
/* For internal usage only */

#include <libregexis024sol/expr_compiler.h>
#include <libregexis024sol/common_codesets.h>
#include <libregexis024fa/finite_automaton.h>
#include <memory>
#include <libregexis024sol/subexpr_fa_transformed.h>
#include <assert.h>
#include <libregexis024fa/selarr_priority_table.h>

namespace regexis024 {
    struct ParsingContext{
        /* Those subexpressions, that are tracket by s`a are forbidden from nesting inside themselves */
        std::vector<bool> is_inside_of_these_sa_subexpressions;
        bool select_cmd_encountered = false;
        RegexPriorityTable priority_table;
        bool dfa_cmd_activated = false;
        /* Completely failing to build dfa with this flag on will result in no error */
        bool dfa_cmd_nonimportant = false;
        /* With this flag, your dfa should be absolutely pure, no forks are allowed. */
        bool dfa_cmd_unforgiving = false;

        /* Reference to active cc set (actually, there is only one cc, but who cares, I placed
         * it here to lower the number of arguments in ParseCall methods, again WHO CARES?) */
        const CommonCodesets& cc;
        explicit ParsingContext(const CommonCodesets& cc_): cc(cc_){}
    };

    typedef REGEX_IS024_MeaningContext ctx_t;
    struct ParseCall;
    typedef std::unique_ptr<ParseCall> chekushka;
    struct ParseCall{
        SubExprCompiled& result;
        explicit ParseCall(SubExprCompiled &result) : result(result) {}
        virtual ~ParseCall() = default;
        virtual chekushka afterReceive(ctx_t& ctx, ParsingContext& pctx, FA_Container& fa) { assert(false); }
        virtual chekushka firstTime(ctx_t& ctx, ParsingContext& pctx, FA_Container& fa) { assert(false); }
    };

    struct TopLvl_ParseCall: public ParseCall{
        explicit TopLvl_ParseCall(SubExprCompiled &result) : ParseCall(result) {}
        chekushka afterReceive(ctx_t &ctx, ParsingContext &pctx, FA_Container &fa) override;
        chekushka firstTime(ctx_t &ctx, ParsingContext &pctx, FA_Container &fa) override;
    };

    struct BracketLvl_ParseCall: public ParseCall{
        /* -1 if this is a normal bracket expression. Otherwise, it is an index in ctx.retrieval_info vector */
        int64_t namedSubexpressionId;
        SubExprCompiled tmp_ret_buff;
        explicit BracketLvl_ParseCall(SubExprCompiled& result, int64_t namedSubexpressionId) :
                ParseCall(result), namedSubexpressionId(namedSubexpressionId) {}
        chekushka afterReceive(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa) override;
        chekushka firstTime(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa) override;
    };

    struct ForkLvl_ParseCall: public ParseCall{
        std::vector<SubExprCompiled> options;
        explicit ForkLvl_ParseCall(SubExprCompiled &result) : ParseCall(result) {}
        chekushka afterReceive(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa);
        chekushka firstTime(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa);
    };

    struct Sequence_ParseCall: public ParseCall{
        std::vector<SubExprCompiled> parts;
        explicit Sequence_ParseCall(SubExprCompiled &result) :ParseCall(result) {}
        chekushka afterReceive(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa);
        chekushka firstTime(REGEX_IS024_MeaningContext& ctx, ParsingContext& pctx, FA_Container& fa);
    };
}

#endif //LIBREGEXIS024_SRC_LIBREGEXIS024SOL_EXPR_PARSE_FUNCTIONS_EPF_H
