#include <libregexis024sol/sol_misc_base.h>
#include <libregexis024vm/utils.h>

namespace regexis024 {
    void report(REGEX_IS024_MeaningContext &ctx, const char *error) {
        if (!ctx.error){
            ctx.error = true;
            ctx.error_msg = error;
        }
    }

    bool isEnd(REGEX_IS024_MeaningContext &ctx) {
        return ctx.pos == ctx.input_size;
    }

    int32_t peep(REGEX_IS024_MeaningContext &ctx) {
        //    printf("pos = %lu\n", ctx.pos);
        if (isEnd(ctx))
            return -1;  // This is probably the only place where getting negative return does not generate error
        int32_t cp; size_t sz;
        utf8_string_iterat(cp, sz, ctx.pos, ctx.input, ctx.input_size);
        if (cp < 0)
            report(ctx, "encoding error");
        return cp;
    }

    int32_t readChar(REGEX_IS024_MeaningContext &ctx) {
        //    printf("READ pos = %lu\n", ctx.pos);
        int32_t cp; size_t sz;
        utf8_string_iterat(cp, sz, ctx.pos, ctx.input, ctx.input_size);
        if (cp >= 0)
            ctx.pos += sz;
        else
            report(ctx, "bruh what?? How this even happened");
        return cp;
    }

    bool is_REGEX024_nameConstituent(int32_t ch) {
        return ('0' <= ch && ch <= '9') || ('a' <= ch && ch <= 'z') || ('A' <= ch && ch <= 'Z');
    }

    std::string tryRead_REGEX024_name(REGEX_IS024_MeaningContext &ctx) {
        std::string res;
        while (true){
            int32_t ch = peep(ctx);
            if (is_REGEX024_nameConstituent(ch)){
                res += (char)ch;
                readChar(ctx);
            } else {
                break;
            }
        }
        return res;
    }
}
