#include <libregexis024fa/codeset.h>
#include <libregexis024vm/utils.h>
#include <stdio.h>

using namespace regexis024;

void test_ccs_fnc(const codeset_t &got, const codeset_t &expected){
    static int id = 1;
    if (got == expected) {
        printf("Test %d passed\n", id++);
    } else {
        printf("Test %d failed\n", id);
        std::terminate();
    }
}

void invert_test(const codeset_t& A, const codeset_t& C){
    test_ccs_fnc(invert_set(A), C);
    test_ccs_fnc(invert_set(C), A);
}

void merge_test(const codeset_t& A, const codeset_t& B, const codeset_t& C){
    test_ccs_fnc(merge_sets(A, A), A);
    test_ccs_fnc(merge_sets(B, B), B);
    test_ccs_fnc(merge_sets(A, B), C);
    test_ccs_fnc(merge_sets(B, A), C);
}

void intersect_test(const codeset_t& A, const codeset_t& B, const codeset_t& C){
    test_ccs_fnc(intersect_sets(A, A), A);
    test_ccs_fnc(intersect_sets(B, B), B);
    test_ccs_fnc(intersect_sets(A, B), C);
    test_ccs_fnc(intersect_sets(B, A), C);
}

int main(){
    merge_test({{34, 111}}, {},       {{34, 111}});
    merge_test({{1, 1}}, {{3, 3}},    {{1, 1}, {3, 3}});
    invert_test({{0, 1}},             {{2, UINT32_MAX}});
    invert_test({{32, 34}},         {{0, 31}, {35, UINT32_MAX}});
    merge_test({{10, 10}, {20, 20}}, {{19, 19}},    {{10, 10}, {19, 20}});
    merge_test({{0, 5}, {7, 10}}, {{4, 6}},   {{0, 10}});
    merge_test({{1, 10}, {50, 60}}, {{11, 70}},    {{1, 70}});
    merge_test({{23, 23}, {56, 100}}, {{30, 55}},   {{23, 23}, {30, 100}});
    intersect_test({{100, 200}, {300, 400}}, {},    {});
    intersect_test({{2, 30}}, {{15, 50}},    {{15, 30}});
    intersect_test({{10, 30}}, {{15, 25}},   {{15, 25}});
    intersect_test({{10, 20}}, {{21, 30}},    {});
    intersect_test({{1, 100}, {150, 200}}, {{50, 175}},   {{50, 100}, {150, 175}});
    intersect_test({{1, 100}}, {}, {});
    intersect_test({{50, 50}}, {{50, 50}},   {{50, 50}});
    intersect_test({{49, 49}}, {{50, 50}},   {});
    intersect_test({{1, 20}, {50, 80}}, {{10, 55}, {60, 100}}, {{10, 20}, {50, 55}, {60, 80}});
    merge_test({{2, 3}, {5, 5}, {7, 7}}, {{1, 1}, {3, 7}}, {{1, 7}});
    return 0;
}