#include <libregexis024tools/stringmatching.h>
#include <assert.h>
#include <stdexcept>
#include <libregexis024sol/expr_compiler.h>

using namespace regexis024;
using namespace std;

void test(const string& input, const string& pattern, const MatchInfo& right_answer) {
    MatchInfo given_answer;
    track_var_list retTrackVarList;
    string retStatus;
    int ret = matchStrToRegexp(input, pattern, given_answer, retTrackVarList, retStatus);
    if (ret < 0) {
        throw runtime_error("Test failed. matching. " + retStatus);
    }
    if (given_answer != right_answer) {
        throw runtime_error("Test failed.");
    }
    printf("Test passed\n");
}

int main() {
    test("11aa", "^!A;\\B!A;\\b!any;\\B!any;$", MatchInfo({}, {}));
    test("aa11", "^!A;\\B!A;\\b!any;\\B!any;$", MatchInfo({}, {}));
    test("a111", "^!A;\\B!A;\\b!any;\\B!any;$", MatchInfo());
    test("aa11", "^!A;\\B!A;\\B!any;\\B!any;$", MatchInfo());
    test("1a11", "^!A;\\B!A;\\B!any;\\B!any;$", MatchInfo());
    test("11aa", "!dfa;^!A;\\B!A;\\b!any;\\B!any;$", MatchInfo({}, {}));
    test("aa11", "!dfa;^!A;\\B!A;\\b!any;\\B!any;$", MatchInfo({}, {}));
    test("a111", "!dfa;^!A;\\B!A;\\b!any;\\B!any;$", MatchInfo());
    test("aa11", "!dfa;^!A;\\B!A;\\B!any;\\B!any;$", MatchInfo());
    test("1a11", "!dfa;^!A;\\B!A;\\B!any;\\B!any;$", MatchInfo());
    test("LINE\r\nFirst:Second\r\nThird:12\r\n\r\n",
        "!dfa;!select{fieldname{ca}fieldbody{ca}}^^^LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n$$$",
        MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}, {0, 20}, {1, 25}, {2, 26}, {3, 28}}, {20, 25, 26, 28}));
    test("LINE\r\nFirst:Second\r\nThird:12\r\n\r\n",
        "!dfa;!select{fieldname{ca}fieldbody{ca}}^LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+\\>):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
        MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}, {0, 20}, {1, 25}, {2, 26}, {3, 28}}, {20, 25, 26, 28}));
    test("LINE\r\nFirst:Second\r\nThird:12\r\n\r\n",
        "!dfa;!select{fieldname{ca}fieldbody{ca}}^LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
        MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}, {0, 20}, {1, 25}, {2, 26}, {3, 28}}, {20, 25, 26, 28}));
    test("LINE\r\nFirst:Second\r\n\r\n",
    "!select{fieldname{ca}}LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
    MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}}, {6, 11}));
    test("LINE\r\nFirst:Second\r\n\r\n",
    "!select{fieldname}LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
    MatchInfo({{0, 12}, {1, 18}}, {6, 11}));
    test("LINE\r\nFirst:Second\r\nThird:12\r\n\r\n",
        "!select{fieldname{ca}fieldbody{ca}}LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
        MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}, {0, 20}, {1, 25}, {2, 26}, {3, 28}}, {20, 25, 26, 28}));
    test("абвгд", "абвгд", MatchInfo({}, {}));
    test("абвввввввгд", "абв*г+д", MatchInfo({}, {}));
    test("абвввввввд", "абв*г+д", MatchInfo());
    test("LINE\r\nFirst:Second\r\nThird:12\r\n\r\n",
    "!dfa;^LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
    MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}, {0, 20}, {1, 25}, {2, 26}, {3, 28}}, {}));
    test("LINE\r\nFirst:Second\r\n\r\n",
        "LINE\r\n(#fieldname([\\u0021-\\u007E&^:]+):#fieldbody([\\u0000-\\u007F&^\r\n]*)\r\n)*\r\n",
        MatchInfo({{0, 6}, {1, 11}, {2, 12}, {3, 18}}, {}));
    test("C111111111111", "C\\>1*", MatchInfo({}, {}));
    test("GET / HTTP/1.1\r\nHost: example.com\r\nAAAAA: a\rfaafafdf\r\n\r\n",
        "!dfa;(GET|POST) / HTTP/(1.1|1.0|0.9)\r\n([\\u0021-\\u007E&^:]+:([\\u0000-\\u007F&^\r\n])*\r\n)*\r\n",
        MatchInfo());
    test("\r24234\r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo());
    test("\n3432\r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo());
    test("3:::;;432\r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo({}, {}));
    test("3:::;;432   \r\n", "[\\u0000-\\u007F&^\r\n]*\r\n", MatchInfo({}, {}));
    test("GET / HTTP/0.9\r\nHost: bibur at\r\nContent-type: html\r\n\r\n",
        "^(GET|POST\\>) / HTTP/(1.1|1.0|0.9)\r\n([\\u0021-\\u007E&^:]+:([\\u0000-\\u007F&^\r\n])*\r\n)*\r\n",
        MatchInfo({}, {}));
    test("b", "#boba(b)", MatchInfo({{0, 0}, {1, 1}}, {}));
    test("abc", "!selarr{boba{ca}}^a#boba(b)c$", MatchInfo({{0, 1}, {1, 2}}, {1, 2}));
    for (int i = 0; i < 64; i++) {
        std::string T;
        T += ('a' + (i >> 3));
        T+= ('a' + (i % 8));
        test(T, "(((a|b)|(c|d))|((e|f)|(g|h)))!r{2}", MatchInfo({}, {}));
    }
    test("abba", "!select{M{max}}a#M(b*)a", MatchInfo({}, {1, 3}));
    test("abba", "!dfa;!select{M{max}}a#M(b*)a", MatchInfo({}, {1, 3}));
    test("abba", "!select{M{max}}a#M(!any;*)a", MatchInfo({}, {1, 3}));
    test("abba", "!dfa;!select{M{max}}a#M(!any;*)a", MatchInfo({}, {1, 3}));
    test("", "", MatchInfo({}, {}));
    test("a", "a", MatchInfo({}, {}));
    test("a3", "[abc]3", MatchInfo({}, {}));
    test("b3", "[abc]3", MatchInfo({}, {}));
    test("c3", "[abc]3", MatchInfo({}, {}));
    test("aa", "aa", MatchInfo({}, {}));
    test("aaaaa", "a*", MatchInfo({}, {}));
    test("bababbaa", "[ab]*", MatchInfo({}, {}));
    test("bababbaa", "!dfa;[ab]*", MatchInfo({}, {}));
    return 0;
}
