#include <libregexis024vm/utils.h>

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>

#ifndef __ORDER_LITTLE_ENDIAN__
#error "Big endian is currently unsupported"
#endif

namespace regexis024 {
    int utf8_retrieve_size(char firstByte) {
        if (!((uint8_t)firstByte & 0b10000000))
            return 1;
        uint8_t a = 0b11000000;
        uint8_t b = 0b00100000;
        for (int i = 2; i <= 4; i++){
            if (((uint8_t)firstByte & (a | b)) == a)
                return i;
            a |= b;
            b >>= 1;
        }
        return -1;
    }

    int32_t utf8_retrieve_character(int sz, size_t pos, const char *string) {
        if (sz == 1)
            return (uint8_t)string[pos];
        uint32_t v = (uint8_t)string[pos] & (0b01111111 >> sz);
        pos++;
        for (int i = 1; i < sz; i++){
            uint32_t th = (uint8_t)string[pos];
            if ((th & 0b11000000) != 0b10000000)
                return -1;
            v <<= 6;
            v |= (th & 0b00111111);
            pos++;
        }
        assert(v <= INT32_MAX);
        return static_cast<int32_t>(v);
    }

    void utf8_string_iterat(int32_t &cp, size_t &adj, size_t pos, const char *string, size_t string_size) {
        if (pos >= string_size) {cp = -1; return;}
        adj = utf8_retrieve_size(string[pos]);
        if (adj < 0 || pos + adj > string_size) {cp = -1; return;}
        if ((cp = utf8_retrieve_character(adj, pos, string)) < 0) {cp = -1;}
    }

    bool is_string_in_stringset(const char *strSample, const char **strSet) {
        const char** cmpSubject = strSet;
        while ((*cmpSubject) != NULL){
            if (strcmp(strSample, *cmpSubject) == 0)
                return true;
            cmpSubject++;  // += 8 bytes
        }
        return false;
    }
}