/* This file is used for testing purposes only. Do not copy this file to installation prefix.
 * This tehnique exploits C compiler capabilities to get regex024 assembler for free*/
#ifndef LIBREGEXIS024_BYTE_CODE_ASSEMBLER_H
#define LIBREGEXIS024_BYTE_CODE_ASSEMBLER_H

#include "vibe_check.h"

#include <libregexis024vm/vm_opcodes.h>
#include <string>
#include <vector>
#include <map>
#include <stdio.h>

using namespace regexis024;

struct assembler_context_bookmark{
    near_ptr_t pos_in_r024program;
    int LINE;
};

struct pending_bookmark{
    /* Must fill this byte with pos of pos_in_r024program in assembler_context_bookmark
     * In a sense, this is a pointer to a NULL pointer that is yet to become normal kinda pointer */
    near_ptr_t pos_in_r024program;
    const char* name;
    /* LINE of the reference is needed in case of error */
    int LINE;
};

struct assembler_context{
    std::map<std::string, assembler_context_bookmark> bookmarks;
    std::vector<pending_bookmark> unresolved_references;
    std::vector<uint8_t> result;

    void declare_bookmark(const char* name, int LINE_of_this){
        if (bookmarks.count(name)){
            fprintf(stderr, "Double bookmark '%s' definition in lines %d and %d\n", name, bookmarks[name].LINE, LINE_of_this);
            exit(1);
        }
        bookmarks[name] = {result.size(), LINE_of_this};
    }

    void resolve_references(){
        for (pending_bookmark& br: unresolved_references){
            if (bookmarks.count(br.name) == 0){
                fprintf(stderr, "Unknown bookmark '%s' is requested on line %d\n", br.name, br.LINE);
                exit(1);
            }
            /* pending bookmerk requests should be added only with beg_for_bookmark method,
             * or else SEGFAULT will be your frequent guest */
            *reinterpret_cast<near_ptr_t *>(&result[br.pos_in_r024program]) = bookmarks[br.name].pos_in_r024program;
        }
    }

    void put_byte(uint8_t x){
        result.push_back(x);
    }

    void put_word(uint16_t x){
        put_byte(x & UINT8_MAX);
        put_byte(x >> 8);
    }

    void put_doubleword(uint32_t x){
        put_word(x & UINT16_MAX);
        put_word(x >> 16);
    }

    void put_quadword(uint64_t x){
        put_doubleword(x & UINT32_MAX);
        put_doubleword(x >> 32);
    }

    void beg_for_bookmark(const char* name, int LINE_of_this){
        unresolved_references.push_back({result.size(), name, LINE_of_this});
        put_quadword(0);
    }
};

#define msh_put_instr(ename)          daCtx.put_byte(regex024_opcodes::ename);
#define msh_put_sslot(ssid)           daCtx.put_doubleword(ssid);
#define msh_put_track_arr_ind(i)      daCtx.put_word(i);
#define msh_put_x(meth, x)            daCtx.meth(x);
/* Here my assembler begs for bookmark to jump on */
#define msh_bookmark_reference(name)  daCtx.beg_for_bookmark(name, __LINE__);

#define s_BEGIN_ASSEMBLER_CONTEXT() { assembler_context daCtx{};
#define s_END_ASSEMBLER_CONTEXT(pass_vec) daCtx.resolve_references(); (pass_vec) = std::move(daCtx.result);  }

/*  Here user declares a bookmark */
#define c_BOOKMARK(name) daCtx.declare_bookmark((name), __LINE__);

#define i_READ(ssid)                    msh_put_instr(READ)       msh_put_sslot(ssid)
#define i_READZ()                       msh_put_instr(READZ)
#define i_JUMP(bookmark)                msh_put_instr(JUMP)       msh_bookmark_reference(bookmark)

#define msh_conditional_jump(condition, size_postfix, meth, x, bookmark) \
                msh_put_instr(JC ## condition ## _ ## size_postfix) \
                    msh_put_x(meth, x)          msh_bookmark_reference(bookmark)

#define i_JCEQUAL_B(x, bookmark)   msh_conditional_jump(EQUAL, B, put_byte, x, bookmark)
#define i_JCEQUAL_W(x, bookmark)   msh_conditional_jump(EQUAL, W, put_word, x, bookmark)
#define i_JCEQUAL_DW(x, bookmark)   msh_conditional_jump(EQUAL, DW, put_doubleword, x, bookmark)
#define i_JCEQUAL_QW(x, bookmark)   msh_conditional_jump(EQUAL, QW, put_quadword, x, bookmark)

#define i_JCLESS_B(x, bookmark)   msh_conditional_jump(LESS, B, put_byte, x, bookmark)
#define i_JCLESS_W(x, bookmark)   msh_conditional_jump(LESS, W, put_word, x, bookmark)
#define i_JCLESS_DW(x, bookmark)   msh_conditional_jump(LESS, DW, put_doubleword, x, bookmark)
#define i_JCLESS_QW(x, bookmark)   msh_conditional_jump(LESS, QW, put_quadword, x, bookmark)

#define i_JCGRTR_B(x, bookmark)   msh_conditional_jump(GRTR, B, put_byte, x, bookmark)
#define i_JCGRTR_W(x, bookmark)   msh_conditional_jump(GRTR, W, put_word, x, bookmark)
#define i_JCGRTR_DW(x, bookmark)   msh_conditional_jump(GRTR, DW, put_doubleword, x, bookmark)
#define i_JCGRTR_QW(x, bookmark)   msh_conditional_jump(GRTR, QW, put_quadword, x, bookmark)

#define i_FORK(ssid, bookmark)          msh_put_instr(FORK)   msh_put_sslot(ssid)  msh_bookmark_reference(bookmark)
#define i_MATCH()                       msh_put_instr(MATCH)
#define i_DIE()                         msh_put_instr(DIE)
#define i_PARAM_READ_SS_NUMBER(ssid)      msh_put_instr(PARAM_READ_SS_NUMBER)   msh_put_sslot(ssid)
#define i_PARAM_FORK_SS_NUMBER(ssid)      msh_put_instr(PARAM_FORK_SS_NUMBER)   msh_put_sslot(ssid)
#define i_PARAM_SELARR_LEN(tai)           msh_put_instr(PARAM_SELARR_LEN)       msh_put_track_arr_ind(tai)
#define i_PARAM_COLSIFTFUNC_SET(bookmark) msh_put_instr(PARAM_COLSIFTFUNC_SET)  msh_bookmark_reference(bookmark)
#define i_PARAM_COLSIFTFUNC_WIPE()        msh_put_instr(PARAM_COLSIFTFUNC_WIPE)
#define i_MSG_MULTISTART_ALLOWED(is_allowed)    msh_put_instr(MSG_MULTISTART_ALLOWED)  daCtx.put_byte(is_allowed);
#define i_MSG_FED_INPUT_EXTENDED(left, right, part)   msh_put_instr(MSG_FED_INPUT_EXTENDED) \
                                        daCtx.put_byte(left);   daCtx.put_byte(right);  msh_put_sslot(part)
#define i_DMOV_RABX_SELARR(tai)           msh_put_instr(DMOV_RABX_SELARR)       msh_put_track_arr_ind(tai)
#define DDIST_RABX_SELARR(tai_beg, tai_end)    msh_put_instr(DDIST_RABX_SELARR) \
                                                    msh_put_track_arr_ind(tai_beg)  msh_put_track_arr_ind(tai_end)
#define i_SIFTPRIOR_MIN_RABX()                msh_put_instr(SIFTPRIOR_MIN_RABX)
#define i_SIFTPRIOR_MAX_RABX()                msh_put_instr(SIFTPRIOR_MAX_RABX)
#define i_SIFT_DONE()                         msh_put_instr(SIFT_DONE)
#define i_MOV_COLARR_IMM(tai, qw_x)           msh_put_instr(MOV_COLARR_IMM)     msh_put_track_arr_ind(tai) \
                                                                                    daCtx.put_quadword(qw_x);
#define i_MOV_COLARR_BTPOS(tai)               msh_put_instr(MOV_COLARR_BTPOS)   msh_put_track_arr_ind(tai)
#define i_MOV_SELARR_IMM(tai, qw_x)           msh_put_instr(MOV_SELARR_IMM)     msh_put_track_arr_ind(tai) \
                                                                                    daCtx.put_quadword(qw_x);
#define i_MOV_SELARR_CHPOS(tai)               msh_put_instr(MOV_SELARR_CHPOS)
#define i_INIT()                              msh_put_instr(INIT)
#define i_THROW()                              msh_put_instr(THROW)
#define i_DEBUG()                              msh_put_instr(DEBUG)

#endif //LIBREGEXIS024_BYTE_CODE_ASSEMBLER_H
